/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

package gprobocode;

import java.io.*;
import java.net.*;
import java.util.*;
import java.security.*;
import java.lang.reflect.*;

import javax.swing.*;

import ec.*;
import ec.simple.*;
import ec.gp.*;
import ec.gp.koza.*;
import ec.util.*;

import robocode.*;
import robocode.control.*;
import robocode.util.*;
import robocode.battle.*;
import robocode.manager.*;
import robocode.security.*;
import robocode.manager.ImageManager;
import robocode.peer.robot.RobotStatistics;

/**
 * <p>Title: GPRobocode</p>
 * <p>Description: An adapting interface between Robocode and ECJ</p>
 * <p>Universita` degli Studi di Pavia</p>
 * @author Marco Piastra
 * @version 2003-5-30
 */

public class GPRobocode extends GPProblem implements RobocodeListener,
        SimpleProblemForm {

    static final long serialVersionUID = -4759980263661979669L;

    public final static String P_TANK_PROXY = "gproboproxy.GPTankProxy";

    public final static String P_BATTLE_NUMROUNDS =
        "robocode.battle.numRounds";
    public final static String P_BATTLE_GUNCOOLINGRATE =
        "robocode.battle.rules.gunCoolingRate";
    public final static String P_BATTLE_WIDTH = "robocode.battleField.width";
    public final static String P_BATTLE_INACTIVITY =
        "robocode.battle.rules.inactivityTime";
    public final static String P_BATTLE_SELECTED =
        "robocode.battle.selectedRobots";
    public final static String P_BATTLE_HEIGHT =
        "robocode.battleField.height";

    protected long gameCounter = 0;
    protected BattleProperties properties;
    protected Object robot;
    protected Properties output;

    transient private static RobocodeManager manager;
    transient private int score;

    /**
     * Specializes GPProblem's setup method
     */
    public void setup(final EvolutionState state, final Parameter base) {

        // very important, remember this
        super.setup(state, base);

        manager = robocodeSetup();

        manager.setListener(this);

        // Set battle parameters
        properties = new BattleProperties();

        properties.setGunCoolingRate(state.parameters.getDoubleWithDefault(base.push(P_BATTLE_GUNCOOLINGRATE),
                null, 0.1));
        properties.setNumRounds(state.parameters.getIntWithDefault(base.push(P_BATTLE_NUMROUNDS),
                null, 3));
        properties.setBattlefieldWidth(state.parameters.getIntWithDefault(base.push(P_BATTLE_WIDTH),
                null, 800));
        properties.setInactivityTime(state.parameters.getLongWithDefault(base.push(P_BATTLE_INACTIVITY),
                null, 450));
        properties.setSelectedRobots(state.parameters.getStringWithDefault(base.push(P_BATTLE_SELECTED), "simple.Crazy")
                                     + "," + P_TANK_PROXY);
        properties.setBattlefieldHeight(state.parameters.getIntWithDefault(base.push(P_BATTLE_HEIGHT),
                null, 800));
    }

    /**
     * Specializes GPProblem's evaluation method
     */
    public void evaluate(final EvolutionState state, final Individual ind,
                         final int threadnum) {

        Robot robot = null;
        Properties results = null;

        // Force re-invoking robocodeSetup after resuming a checkpoint
        if (manager == null) {
            manager = robocodeSetup();

            manager.setListener(this);
        }

        // The fitness function adopted is somewhat random  (i.e. due to
        // random initial tank positioning). Hopefully, the re-evaluation
        // of individuals should mitigate these random effects.
        if (!ind.evaluated) {

            // Reset game
            gameCounter++;

            state.output.message("");
            state.output.message("Generation: " + state.generation);
            state.output.message("Evaluation game n. " + gameCounter);

            // Play the game
            manager.getBattleManager().startNewBattle(properties, false);

            // Robcode wait for me, please
            manager.getBattleManager().pauseBattle();

            // Wait until robots are loaded
            try {
                while (!manager.getBattleManager().getBattle().isRobotsLoaded()) {
                    Thread.sleep(100);
                }
            }
            catch (InterruptedException e) {
                e.printStackTrace();
            }

            try {

                // Serialize individual
                ByteArrayOutputStream out = new ByteArrayOutputStream();
                ObjectOutputStream outs = new ObjectOutputStream(out);

                outs.writeObject(ind);
                outs.flush();

                outs.close();
                out.close();

                // Upload to GPTankProxy
                int battleSize = manager.getBattleManager().getBattle().getRobots().size();
                robot =
                    manager.getBattleManager().getBattle().getRobots().elementAt(battleSize - 1).getRobot();

                if (!robot.getClass().getName().equals(P_TANK_PROXY)) {
                    throw new RuntimeException("Wrong robot: "
                                               + robot.getClass().getName());
                }

                Method method = robot.getClass().getMethod("uploadIndividual",
                                                           new Class[] {
                    byte[].class
                });

                method.invoke(robot, new Object[] {
                    out.toByteArray()
                });

                state.output.message("Individual uploaded.");
            }
            catch (Exception e) {
                e.printStackTrace();
            }

            // We're back in business
            manager.getBattleManager().resumeBattle();
            state.output.message("Battle started.");

            // Wait until battle is running
            synchronized (manager) {
                try {
                    manager.wait();
                }
                catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }

            try {

                // Download properties from GPTankProxy, if any
                Method method = robot.getClass().getMethod("downloadOutput",
                                                           new Class[0]);

                output = (Properties) method.invoke(robot, new Object[0]);

                if (output != null) {
                    ByteArrayOutputStream out = new ByteArrayOutputStream();
                    output.store(out, "Output properties");

                    state.output.message(out.toString());
                }
            }
            catch (Exception e) {
                e.printStackTrace();
            }

            RobotStatistics opponentStat =
                (RobotStatistics) manager.getBattleManager().getBattle().getRobots().elementAt(0).getStatistics();
            RobotStatistics gpTankProxyStat =
                (RobotStatistics) manager.getBattleManager().getBattle().getRobots().elementAt(1).getStatistics();

            // An ideal individual must definitely beat its opponent
            score =
                (int) Math.max((opponentStat.getTotalScore() + (properties.getNumRounds() * 100))
                               - gpTankProxyStat.getTotalScore(), 0);

            // Better if all rounds are won
            score +=
                (properties.getNumRounds() - gpTankProxyStat.getTotalFirsts())
                * 100;

            // Compute a final score, somehow
            state.output.message("Battle over (score: " + score + ")");

            // The fitness better be KozaFitness!
            KozaFitness f = ((KozaFitness) ind.fitness);

            f.setStandardizedFitness(state, (float) score);

            // The evaluation of individuals is often a bit of aleatory:
            // random initial positions may in fact yield strange results
            // By forcing the re-evaluation of individuals, this phenomenon
            // might be contained
            // ind.evaluated = true;
        }
    }


    /* begin RobocodeListener interface */

    public void battleComplete(BattleSpecification bs,
                               RobotResults[] results) {

        // Ok, the battle is over: wake up
        synchronized (manager) {
            manager.notifyAll();
        }
    }

    public void battleAborted(BattleSpecification bs) {

        // This would alter the evolutionary run
        System.out.println("*** Battle aborted:");
        System.out.println("    evolutionary run must be aborted too.");

        System.exit(1);
    }

    public void battleMessage(String s) {
        System.out.println(s);
    }

    /* end RobocodeListener interface */


    /**
     * Setup of the robocode evaluation environment
     */
    public static RobocodeManager robocodeSetup() {
        RobocodeManager manager = null;

        try {
            manager = new RobocodeManager(false, null);

            // Set native look and feel
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());

            Constants.setWorkingDirectory(new File(System.getProperty("user.dir")));

            Thread.currentThread().setName("Application Thread");

            // This allows us, in Java 1.4, to use a custom 'classpath' policy.
            RobocodeSecurityPolicy securityPolicy =
                new RobocodeSecurityPolicy(Policy.getPolicy());

            Policy.setPolicy(securityPolicy);

            boolean securityOn = false;

            System.setSecurityManager(new RobocodeSecurityManager(Thread.currentThread(),
                    manager.getThreadManager(), false));
            RobocodeFileOutputStream.setThreadManager(manager.getThreadManager());

            ThreadGroup tg = Thread.currentThread().getThreadGroup();

            while (tg != null) {
                ((RobocodeSecurityManager) System.getSecurityManager()).addSafeThreadGroup(tg);

                tg = tg.getParent();
            }

            SecurePrintStream sysout = new SecurePrintStream(System.out,
                    true, "System.out");
            SecurePrintStream syserr = new SecurePrintStream(System.err,
                    true, "System.err");
            SecureInputStream sysin = new SecureInputStream(System.in,
                    "System.in");

            System.setOut(sysout);
            System.setErr(syserr);
            System.setIn(sysin);

            File robots = new File(Constants.cwd(), "robots");

            if (!robots.exists() ||!robots.isDirectory()) {
                System.err.println(new File(Constants.cwd(), "").getAbsolutePath()
                                   + " is not a valid directory to start Robocode in.");
                System.exit(8);
            }

            manager.getWindowManager().showSplashScreen();
            manager.getWindowManager().showRobocodeFrame();
            manager.getWindowManager().getRobocodeFrame().setState(JFrame.ICONIFIED);

            try {

                // Let robocode complete its own startup
                Thread.sleep(1000);
            }
            catch (InterruptedException e) {
                e.printStackTrace();
            }

            return manager;
        }
        catch (Throwable e) {
            e.printStackTrace();

            return null;
        }
    }

    public synchronized void waitForBattleComplete() {
        try {
            System.out.println("Entering wait: " + this);
            wait();
            System.out.println("Exiting wait: " + this);
        }
        catch (InterruptedException e) {
            e.printStackTrace();
        }
    }

    public synchronized void notifyBattleComplete() {
        notifyAll();
        System.out.println("Notified: " + this);
    }

    /**
     * Just for testing
     */
    public static void main(String[] args) {

        RobocodeManager manager = robocodeSetup();

    }

}

